@@ -8,6 +8,7 @@ import pandas as pd
 import sys
 
 import pyjava.utils as utils
+import requests
 from pyjava.serializers import ArrowStreamSerializer
 from pyjava.serializers import read_int
 from pyjava.utils import utf8_deserializer
@@ -29,33 +30,27 @@ class DataServer(object):
 class LogClient(object):
     def __init__(self, conf):
         self.conf = conf
-        if 'spark.mlsql.log.driver.host' in self.conf:
-            self.log_host = self.conf['spark.mlsql.log.driver.host']
-            self.log_port = self.conf['spark.mlsql.log.driver.port']
+        if 'spark.mlsql.log.driver.url' in self.conf:
+            self.url = self.conf['spark.mlsql.log.driver.url']
             self.log_user = self.conf['PY_EXECUTE_USER']
             self.log_token = self.conf['spark.mlsql.log.driver.token']
             self.log_group_id = self.conf['groupId']
-            import socket
-            self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            self.conn.connect((self.log_host, int(self.log_port)))
-            buffer_size = int(os.environ.get("SPARK_BUFFER_SIZE", 256))
-            self.infile = os.fdopen(
-                os.dup(self.conn.fileno()), "rb", buffer_size)
-            self.outfile = os.fdopen(
-                os.dup(self.conn.fileno()), "wb", buffer_size)
 
     def log_to_driver(self, msg):
-        if not self.log_host:
-            print(msg)
+        if 'spark.mlsql.log.driver.url' not in self.conf:
+            if self.conf['PY_EXECUTE_USER'] and self.conf['groupId']:
+                logging.info("[owner] [{}] [groupId] [{}] __MMMMMM__ {}".format(self.conf['PY_EXECUTE_USER'],
+                                                                                self.conf['groupId'], msg))
+            else:
+                logging.info(msg)
             return
-        from pyjava.serializers import write_bytes_with_length
         import json
         resp = json.dumps(
             {"sendLog": {
                 "token": self.log_token,
-                "logLine": "[owner] [{}] [groupId] [{}] {}".format(self.log_user, self.log_group_id, msg)
+                "logLine": "[owner] [{}] [groupId] [{}] __MMMMMM__ {}".format(self.log_user, self.log_group_id, msg)
             }}, ensure_ascii=False)
-        write_bytes_with_length(resp, self.outfile)
+        requests.post(self.url, data=resp, headers={'content-type': 'application/x-www-form-urlencoded;charset=UTF-8'})
 
     def close(self):
         if hasattr(self, "conn"):
@@ -182,6 +177,7 @@ class PythonProjectContext(object):
 
 class RayContext(object):
     cache = {}
+    conn_cache = {}
 
     def __init__(self, python_context):
         self.python_context = python_context
@@ -191,14 +187,18 @@ class RayContext(object):
         self.is_dev = utils.is_dev()
         self.is_in_mlsql = True
         self.mock_data = []
-        for item in self.python_context.fetch_once_as_rows():
-            self.server_ids_in_ray.append(str(uuid.uuid4()))
-            self.servers.append(DataServer(
-                item["host"], int(item["port"]), item["timezone"]))
+        if "directData" not in python_context.conf:
+            for item in self.python_context.fetch_once_as_rows():
+                self.server_ids_in_ray.append(str(uuid.uuid4()))
+                self.servers.append(DataServer(
+                    item["host"], int(item["port"]), item["timezone"]))
 
     def data_servers(self):
         return self.servers
 
+    def conf(self):
+        return self.python_context.conf
+
     def data_servers_in_ray(self):
         import ray
         from pyjava.rayfix import RayWrapper
@@ -248,8 +248,13 @@ class RayContext(object):
         if url is not None:
             from pyjava.rayfix import RayWrapper
             ray = RayWrapper()
-            ray.shutdown()
-            ray.init(url, **kwargs)
+            is_udf_client = context.conf.get("UDF_CLIENT")
+            if is_udf_client is None:
+                ray.shutdown()
+                ray.init(url, **kwargs)
+            if is_udf_client and url not in RayContext.conn_cache:
+                ray.init(url, **kwargs)
+                RayContext.conn_cache[url] = 1
         return context.rayContext
 
     def setup(self, func_for_row, func_for_rows=None):
@@ -294,10 +299,13 @@ class RayContext(object):
             for row in RayContext.fetch_once_as_rows(shard):
                 yield row
 
-    def fetch_as_dir(self, target_dir):
-        if len(self.data_servers()) > 1:
+    def fetch_as_dir(self, target_dir, servers=None):
+        if not servers:
+            servers = self.data_servers()
+        if len(servers) > 1:
             raise Exception("Please make sure you have only one partition on Java/Spark Side")
-        items = self.collect()
+
+        items = RayContext.collect_from(servers)
         streaming_tar.save_rows_as_file(items, target_dir)
 
     def build_result(self, items, block_size=1024):
@@ -306,6 +314,12 @@ class RayContext(object):
     def build_result_from_dir(self, target_path):
         self.python_context.build_result_from_dir(target_path)
 
+    @staticmethod
+    def parse_servers(host_ports):
+        hosts = host_ports.split(",")
+        hosts = [item.split(":") for item in hosts]
+        return [DataServer(item[0], int(item[1]), "") for item in hosts]
+
     @staticmethod
     def fetch_as_repeatable_file(context_id, data_servers, file_ref, batch_size):
         import pyarrow as pa
